"""
A skeleton class to store data related to Jacobi Forms
Author: David Gruenewald 2006-8-30, updated/corrected/extended by Nathan 
Ryan 2007-6-21
"""

class JacobiForm:

    ## Initialise the Jacobi form
    def __init__(self, weight, index, formexpansion):
        self.__weight = weight
        self.__index = index
        self.__expansion = formexpansion
        self.__precision = formexpansion.prec()

    def __add__(self, other):
        """
        Returns the sum of the self and other (componentwise)
        """
        assert (self.__weight == other.weight()) and (self.__index == other.index())
        expansion = self.__expansion + other.expansion()
        return JacobiForm(self.__weight,  self.__index, expansion)
    
    # Some accessor functions 
    def __repr__(self):
        return self.__expansion.__repr()

    def weight(self):
        return self.__weight
 
    def index(self):
        return self.__index
 
    def expansion(self):
        return self.__expansion

    def precision(self):
        return self.__precision

## End Jacobi form class

def skoruppa_isomorphism(f,g, precision=Integer(20)):
    r"""
    Calculates the image of the pair of modular forms (f,g) under
    Skoruppa's isomorphism I : M_{k} \oplus S_{k+2} -> J_{k,1}

    Output is a power series \sum_{r^2-4n \leq 0} c(n,r) q^n zeta^r
    with r \in {0,1} and precision in q at least O(q^formprecision)
    """

    k = f.weight()
    assert (k + 2 == g.weight()) | (f == 0) | (g == 0), "incorrect weights!"
    assert (g.q_expansion(1) == 0), "second argument is not a cusp form!"
    Rz = PolynomialRing(QQ,'zeta'); zeta = Rz.gen()
    PS = PowerSeriesRing(Rz, 'q'); q = PS.gen()

    etapow = dedekind_eta_quasi_expansion(precision)^(-6)
    
    ## create the Jacobi forms A=a*etapow and B=b*etapow in stages
    ## r is the exponent of zeta^r and s is the exponent of q^((s^2+r^2-1)/4)
    ## note that we must have r-s = 1 mod 2 always.
    ## compute only r values in {0,1} corresponding to discriminants 0,1 mod 4

    ## general term of a is s^2*(-1)^r*q^((s^2+r^2-1)/4)*zeta^r
    ## general term of b is (-1)^r*q^((s^2+r^2-1)/4)*zeta^r
    ## s runs over ZZ but fortunately the terms are even in s


    (a,b) = (0,0)
    # the following computes a to precision at least O(q^precision)
    for s in range(1, precision):
        if (mod(s,2) == 0):  ## then r = 1
            a = a - 2 * s^2 * q^(s^2/4) * zeta
            b = b - 2 * q^(s^2/4)*zeta
        else:  ## then r = 0 
            a = a + 2 * s^2 * q^((s^2 - 1) / 4)
            b = b + 2 * q^((s^2 - 1) / 4)

    ## also have <s=0, r=1> term to add to b:
    b = b - zeta
    ## same precision for b.


    ## The sum defining A (resp. B) has even powers of s, so I'll have a loop
    ## that goes from 1 to precision (I'll add the s = 0 term in by
    ## hand). If s is even, I will loop through all odd r's and if 
    ## s is odd, I'll loop through all even r's in order to 
    ## determine the r,sth coefficient of A (resp. B).

    (A, B) = (a *  etapow, b * etapow)

    ## calculate the image of the pair of modular forms (f,g) under
    ## Skoruppa's isomorphism I : M_{k} \oplus S_{k+2} -> J_{k,1}\

    fderiv = PS(f.qexp(precision).derivative())
    
    ( f, g ) = ( PS(f.qexp(precision)), PS(g.qexp(precision)) )

    Ifg = k/2 * f * A - q * fderiv * B + g * B

    return JacobiForm(k, 1, Ifg)    

## end skoruppa isomorphism
        
def dedekind_eta_quasi_expansion(prec=Integer(20)):
    """
    returns the infinite product over n>=1 of (1-q^n),
    truncated according to the given precision prec.
    This would be the Dedekind eta q-expansion if
    multiplied by q^(1/24).
    """

    ## needs to be heavily optimized!!!! (how...)

    R = PowerSeriesRing(QQ,'q'); q = R.gen()
    return prod([R(Integer(1)-q**n + O(q**prec)) for n in range(Integer(1),prec)])
## end dedekind eta

def reduce_GL(a,b,c):
    """
    Returns the GL2(ZZ)-reduced form equivalent to this (positive semidefinite)
    quadratic form.
    """
    ## Adapting algorithm 5.4.2 found in Cohen's
    ## "Computational Algebraic Number Theory".
    
    b = abs(b)
    while not (b <= a and a<= c):  ## while not GL-reduced
        twoa = 2 * a
        if b not in range(-a+1, a+1): 
            ## apply Euclidean step
            q = b // twoa
            r = b % twoa
            if r > a:
                ## want r to be such that -a < r <= a
                r = r - twoa
                q = q + 1
            c = c - b * q + a * q^2
            b = r
        if a > c:
            (a,c) = (c,a)
        b = abs(b)
    ## end while
    return (a,b,c)
## end GL2-reduce

def maass_lift(J):
    """
    Returns the Maass lift of a Jacobi form of weight k, index 1.
    ie, a genus two Siegel modular form for the full modular group
    upto precision
    """
    assert J.index() == 1
    k = J.weight()

    ## First put the coefficients into a dictionary:
    ## J.expansion().list()[i].list()[j] gives the coefficient of q^i*zeta^j
    
    Cphi = {0 : 0} ## initialize dictionary, value changes in loop if needed
    qcoeffs = J.expansion().list()


    Rz = PolynomialRing(QQ,'zeta'); zeta = Rz.gen()



    for i in range(len(qcoeffs)):
        qi = qcoeffs[i]
        ## Need to edit this line, need to figure this out... NCR
        zetacoeffs = Rz(qi).list()
        if len(zetacoeffs) == 0:
            zetacoeffs.append(0)
            zetacoeffs.append(0)
        if len(zetacoeffs) == 1: zetacoeffs.append(0)
        for j in range(2):  ## of which there are 2
            coeffij = zetacoeffs[j]
            Cphi[j^2 - 4*i] = coeffij 
            

    ## so we can access the coefficient corresponding to
    ## discriminant D by going Chi[D] - brilliant SAGE!

    ## the most negative discriminant occurs when i is largest
    ## and j is zero.  That is, discriminant 0^2-4*i
    ## Note that i < J.precision()

    maxD = -4*i

    """
    Return some maass coefficients of the resulting Siegel
    modular form of weight k=Weight(f).
    """

    ## The constant term is given by -Cphi[0]*B_{2k}/(4*k)
    ## (note in Skoruppa's article this coeff has typos)
    ## Explain?
    ## For nonconstant terms,
    ## The Siegel coefficient of q^n * zeta^r * qdash^m is given
    ## by the formula  \sum_{a|gcd(n,r,m)} Cphi[D/a^2] where
    ## D=r^2-4*n*m is the discriminant.
    ## Hence in either case the coefficient
    ## is fully deterimined by the pair (D,gcd(n,r,m))
    ## Put it in a dictionary (hash table)

    ## first, include the coeff corresponding to <0,0,0>:


    maassc = {(0,0): -bernoulli(k)/(2*k)*Cphi[0]}
    
    ## We can compute coeffs of q^n zeta^r qdash^m for which
    ## |r^2-4nm| < |maxD|

    ## Calculate the discriminant-zero maass coefficients
    absmaxD = -maxD
    cphi0 = Cphi[0]
    
    for i in range(1, absmaxD + 1):
        maassc[(0,i)] = sigma(i, k - 1) * cphi0
 
    ## Calculate maass coefficients corresponding to positive definite matrices
    ## ie negative discriminants.
    for disc in [ d for d in range(maxD,0) if is_fundamental_discriminant(d)]:
        for s in range(Integer(1),floor(sqrt(maxD//disc))+1):
            ## add (disc*s^2,t) as a hash key, for each t that divides s
            for t in divisors(s):
                maassc[(disc*s**2,t)] = sum([a**(k-Integer(1))*Cphi[disc*s**2/a**2] for a in divisors(t)])

    ## compute some Siegel coefficients:

    siegelq = dict()

    """
    store_only_reduced -- (bool; default: False) whether or not to store
            coefficients corresponding to reduced triples only. Note that for
            x=(a,b,c) we have coeff(x) = coeff(reduction of x) which allows us to
            make space saving factor of 5/6.  The benefits of storing only reduced
        is(?) outweighed by the higher cost of computing products of expansions.
    """



    """
    for n in range(-maxD):
        for r in range(-maxD):
            ## note that in general there are SMF's for which coeff(n,r,m) doesn\
't equal coeff(n,-r,m)
            ## ( a_(n,r,m)= (\pm 1)^k * a_(n,-r,m) ) but for maass lifts of this\
 kind the coeffs are equal.
            for m in range(-maxD):
                D=r^2-4*m*n
                if D<=0 and D>=maxD :
                    g=gcd([n,r,m])
                    siegelq[(n,r,m)] = maassc[(D,g)]
                    if r <> 0:
                        siegelq[(n,-r,m)] = maassc[(D,g)]
                    triple = (n,r,m)
                    ##print "disc ", D, ", gcd ", g, ", ", triple, ": " , maassc\
[(D,g)]
    """

    # add reduced coefficients to the Siegel dictionary

    maxtrace = RR(sqrt(-maxD/3)).floor()
    """
    bmax = amax
    cmax = amax
    """
    for r in range(0, maxtrace+1):
        for n in range(r, maxtrace+1):
            for m in range(n, maxtrace+1):
                ## loop ranges guarrantee that 0<=r<=n<=m so r^2<=nm<=4nm hence
                ## (n,r,m) is a GL-reduced semipositive definite binary quadratic form.
                if m + n <= maxtrace:
                    D = r^2-4*m*n
                    g = gcd([n,r,m])
                    siegelq[(n,r,m)] = maassc[(D,g)]

    return SiegelGenus2Expansion(siegelq,(maxtrace,maxtrace,maxtrace))

##end of maass_lift

class SiegelGenus2Expansion:

    r"""
    Fourier expansions of genus 2 Siegel modular forms on the \Sp_4(ZZ)
    """

    def __init__(self, coeffs,precision=(Integer(1),Integer(1),Integer(1))):
        """
        Initialise the Siegel expansion given a coefficient dictionary coeffs
        with triples representing quadratic forms as keys

        By precision (a,b,c) we mean having all reduced forms (x,y,z) such
        that (x,y,z) < (a,b,c) componentwise
        """
        (a,b,c) = precision
        discs = [b^2 - 4*a*c for (a,b,c) in coeffs.keys()]
        assert max(discs) <= Integer(0), "non-holomorphic!"
        assert (0, 0, 0) in coeffs.keys()
        self.__coeffdict = coeffs
        self.__precision = precision


    ########
    """ DG:
    You may wonder why there is no instance variable storing the weight of the form.    
    I could add it, but I'm not using it and more importantly, it's not structurally
    required in my opinion.  Ideally an instance of this class would be created in a
    SiegelModularForm class, and in this class you'd have a weight instance variable.
    """
    ########

    def __add__(self, S):
      raise NotImplementedError, 'addition will be implemented in the near future..'
    def __repr__(self):
        orderedkeys = sorted(self.__coeffdict.keys(),tripleorder) ## can use tripleorder as comparison function instead of the default
        ## print first term
        if len(orderedkeys) == Integer(0):
            return "0"
        output = str(self.__coeffdict[orderedkeys[0]]) + '*q^' + str(orderedkeys[0])
        for key in orderedkeys[1:]:
            if self.__coeffdict[key] <> Integer(0) and key[0] <= self.__precision[0] and key[1] <= self.__precision[1] and key[2] <= self.__precision[2]:
                exp = str(key)
                coeff = str(self.__coeffdict[key])
                term = coeff + '*q^' + exp
                output = output + ' + ' + term
        return output


    def __mult(self, S, tup):
        """
        returns coefficients indexed by reduced forms q^(a,b,c) with
        (a,b,c) < tup=(amax,bmax,cmax) componentwise, of the product of the
        fourier expansion of our SMF and the expansion of another SMF called S
        """
        (amax, bmax, cmax) = tup
        s1 = self.__coeffdict
        s2 = S.dict()
        result = dict()
        if len(s1) > len(s2):
            (s1,s2) = (s2,s1)
        for x in s1.keys():
            if x in s2.keys():
                ## appears to be precision friendly - try and calculate coeff
                try:
                    coeff = self.__mult_coeff(x,S)
                except KeyError:
                    ## precision is too high.  Try another form
                    pass
                else:
                    result[x] = coeff
        return result
    ##end def

    def __pow__(self, n):
        """
        returns the expansion to the nth power, only allowing positive integer
        powers
        """
        assert n in ZZ and n >= 0, "the exponent needs to be a nonnegative integer"
        if n == 0:
            return 1
        result = self
        for i in range(n-1):
            result = result * self
        return result
     
    
   

    def __scalar_mult(self,alpha):
        """
        returns the SMF multiplied by the rational scalar alpha
        """
        assert alpha in QQ, "alpha is not a rational scalar!"
        alphadict = dict([(k,alpha*v) for k,v in self.__coeffdict.iteritems()])
        return SiegelGenus2Expansion(alphadict,self.__precision)


    def __mul__(self,S):
        if S in QQ:
            ## do scalar multiplication
            return self.__scalar_mult(S)
        (a1,b1,c1) = self.__precision
        (a2,b2,c2) = S.precision()
        multprecision = (min(a1,a2), min(b1,b2), min(c1,c2))
        multcoeffs = self.__mult(S, multprecision)
        return SiegelGenus2Expansion(multcoeffs, multprecision)

    def precision(self):
        """
        calculate the amount of precision given by the coefficient dictionary
        """
        return self.__precision

    def dict(self):
        """
        returns the coefficient dictionary
        """
        return self.__coeffdict

    def __mult_coeff(self, tup, S):
        """
        returns the coefficient of q^tup in the product of the form and S.
        """
        (a,b,c) = tup
        assert b^2-4*a*c <= Integer(0) , "form is not semipositive definite!"
        ## reduce the form first so that our search space is smaller
        (a,b,c) = reduce_GL(a,b,c)
        multcoeff = 0
        for a1 in range(a + 1):
            for c1 in range(c + 1):
                B = floor(sqrt(4 * a1 * c1))
                for b1 in range(-B, B + 1):  ## Guarantees that (a1,b1,c1) is positive semidefinite
                    a2 = a - a1
                    b2 = b - b1
                    c2 = c - c1
                    if b2^2 <= 4 * a2 * c2:
                        #####print coeff((a1,b1,c1)),(a1,b1,c1), (a2,b2,c2), coeff((a2,b2,c2),coeffsform2)
                        ## then both (a1,b1,c1) and (a2,b2,c2) are
                        ## positive semidefinite so the monomials exist
                        multcoeff = multcoeff + self.coeff((a1,b1,c1)) * S.coeff((a2,b2,c2))
        return multcoeff
    ##end def

    def coeff(self, (a,b,c)):
        """
        Returns the coefficient of q^(a,b,c)
        If after GL-reducing, the reduced triple is not in the coefficient
        dictionary then an error occurs.
        """
        coeffsform = self.__coeffdict
        assert b^2 - 4* a *c <= 0 ,"form is not semipositive definite!"
        b = abs(b)  ## GL2(ZZ)-reduced
        if coeffsform.has_key((a,b,c)):
            return coeffsform[(a,b,c)]
        ## otherwise GL2(ZZ)-reduce (a,b,c) and try again
        (a,b,c) = reduce_GL(a,b,c)
        if coeffsform.has_key((a,b,c)):
            return coeffsform[(a,b,c)]
        ## otherwise error - precision is too low
        raise KeyError, 'precision is not high enough to extract this coefficient'
    ##end def

    def apply_siegel_operator(self):
        """
        returns a string representing the first few terms of the q-expansion of
        the image of this form under the Siegel operator
        """
        ## We're assuming here that all the coefficients are rational numbers.
        PS = PowerSeriesRing(QQ,'q'); q = PS.gen()
        ## The formula requires the coefficients of q^(a,0,0) which we know
        ## equals the coefficient of q^(0,0,a), which is a reduced form exponent.
        ## We know that coefficients of q^(0,0,c) are accessible for c < cmax.
        cmax = self.__precision[2]
        ## Form the power series:
        coeffs = [self.__coeffdict[(0,0,c)] for c in range(cmax)]
        ps = PS(coeffs) + O(q**cmax)
        return ps
    ##end def


##end Degree 2 expansion class


## unused function -
def tripleorder(t1,t2):
    ## triple comparison function
    (a1,b1,c1) = t1
    (a2,b2,c2) = t2
    if a1*c1-b1^2/4 <> a2*c2-b2^2/4:
        return int(4*a1*c1-b1^2 - 4*a2*c2-b2^2)
    if a1<>a2:
        return int(a1 - a2)
    if b1<>b2:
        return int(b1 - b2)
    return c1 - c2

def compute_generators(trace = 10):
    M4 = ModularForms(1,4)
    M6 = ModularForms(1,6)
    M8 = ModularForms(1,8)
    M10 = ModularForms(1,10)
    M14 = ModularForms(1,14)
    E4 = M4.basis()[0]
    E6 = M6.basis()[0]
    E8 = M8.basis()[0]
    E10 = M10.basis()[0]
    delta = CuspForms(1,12).basis()[0]
    
    precision = ((trace)+1)^2*3//4 + 1


    psi4 = maass_lift(skoruppa_isomorphism(E4, M6(0), precision))
    psi6 = maass_lift(skoruppa_isomorphism(E6, M8(0), precision))
    xi10 = maass_lift(skoruppa_isomorphism(M10(0), delta, precision))
    xi12 = maass_lift(skoruppa_isomorphism(delta, M14(0), precision))


    return psi4, psi6, xi10, xi12

class SiegelModularFormsDegree2:
    def __init__(self, wt = 4, prec = 10):
        self._wt = wt
        self._gens =  compute_generators(prec) # gens is the tuple of generators, accessed with [i], precomputation
        self._dim = None
        self._basis = None

    def _dim_(self):
        f(x) = 1/((1-x^4)*(1-x^6)*(1-x^10)*(1-x^12)) 
        F(x) = f.taylor(x, 0, self._wt+1)
        coeff = F(x).coeff(x^self._wt)
        self._dim = coeff
        return coeff
        
    def __repr__(self):
        return "Space of Siegel modular forms of level 1 of weight %s" %self._wt

    def dimension(self):
        if self._dim:
            return self._dim
        dim = self._dim_()
        self._dim = dim
        return dim

    def _basis_(self):
        wt = self._wt
        L = []
        for a in range(wt/4+1):
            for b in range(wt/6+1):
                for c in range(wt/10+1):
                    for d in range(wt/12+1):
                        if 4*a + 6*b + 10*c + 12*d == wt:
                            L.append((a,b,c,d))

        basis = []
        for tuple in L:
            for i in range(4):
                if tuple[i] <> 0:
                    elt = self._gens[i]**tuple[i]
                    break
            for j in range(i+1,4):
                if tuple[j] <> 0:
                    elt = elt * self._gens[j]**tuple[j]
            basis.append(elt)
        self._basis = basis
        return basis
                      
    def basis(self):
        if self._basis:
            return self._basis
        basis = self._basis_()
        self._basis = basis
        return basis

    def _cuspidal_subspace_(self):
        b = self.basis()
        cuspidal_basis = []
        for f in b:
            if f.dict()[(0,0,0)] == 0:
                cuspidal_basis.append(f)
        return SiegelModularCuspFormsDegree2(self._wt,cuspidal_basis)

    def cuspidal_subspace(self):
        return self._cuspidal_subspace_()

class SiegelModularCuspFormsDegree2:
    def __init__(self, wt, basis):
        self._wt = wt
        self._basis = basis
        self._dim = len(basis)

    def __repr__(self):
        return "Space of Siegel modular cuspforms of level 1 of weight %s" %self._wt
 
    def dimension(self):
        return self._dim

    def basis(self):
        return self._basis
    """
    def _tp_action_(self, p):
        '''
        Takes a Siegel Genus 2 Expansion and determines the action of T(p) on it
        '''
        f = self._basis[0]
        wt = self._wt
        new_coeffs = dict()
        for k in f.dict().keys():
            if k[2]*k[1]*k[0] % p == 0:
                continue
            newk = (p*k[0], p*k[1], p*k[2])
            if not tripleorder(newk,f.dict().keys()[0]):
                continue
            newk2 = (k[0]/p, k[1]/p, k[2]/p)
            if not tripleorder(newk2,f.dict().keys()[0]):
                continue
            newk3 = (k[2]/p, k[1], p*k[0])
            if not tripleorder(newk3,f.dict().keys()[0]):
                continue
            new_coeffs[k] = f.dict()[newk] + p^(2*wt - 3)*f.dict()[newk2] + p^(wt - 2)*f.dict()[newk3]
            for nu in range(p):
                sumk = ((k[0] + k[1]*nu + k[2]*nu^2)/p, k[1] + 2*k[2]*nu, p*k[2])
                if not tripleorder(sumk,f.dict().keys()[0]):
                    break
                new_coeffs[k] = new_coeffs[k] + p^(wt - 2)*f.dict()[sumk]
        return new_coeffs
    """                        

        
        
          
 

    
    

   

## end class
